/*
 * Copyright 2013 websudos ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.websudos.phantom.dsl.crud

import org.scalatest.concurrent.PatienceConfiguration
import org.scalatest.time.SpanSugar._

import com.newzly.util.testing.AsyncAssertionsHelper._
import com.websudos.phantom.Implicits._
import com.websudos.phantom.testing.PhantomCassandraTestSuite
import com.websudos.phantom.tables.{Primitive, Primitives}

class SelectTest extends PhantomCassandraTestSuite {

  implicit val s: PatienceConfiguration.Timeout = timeout(10 seconds)

  override def beforeAll(): Unit = {
    super.beforeAll()
    Primitives.insertSchema()
  }

  "Selecting the whole row" should "work fine" in {
    val row = Primitive.sample
    val rcp =  Primitives.insert
        .value(_.pkey, row.pkey)
        .value(_.long, row.long)
        .value(_.boolean, row.boolean)
        .value(_.bDecimal, row.bDecimal)
        .value(_.double, row.double)
        .value(_.float, row.float)
        .value(_.inet, row.inet)
        .value(_.int, row.int)
        .value(_.date, row.date)
        .value(_.uuid, row.uuid)
        .value(_.bi, row.bi).future() flatMap {
        _ => {
          for {
            a <- Primitives.select.fetch
            b <- Primitives.select.where(_.pkey eqs row.pkey).one
          } yield (a, b)

        }
      }

    rcp successful {
      r => {
        r._1 contains row shouldEqual true

        r._2.isDefined shouldEqual true
        r._2.get shouldEqual row
      }
    }
  }

  "Selecting the whole row" should "work fine with Twitter futures" in {
    val row = Primitive.sample
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).execute() flatMap {
      _ => {
        for {
          a <- Primitives.select.collect()
          b <- Primitives.select.where(_.pkey eqs row.pkey).get
        } yield (a contains row, b.get === row)

      }
    }

    rcp successful {
      r => {
        assert(r._1)
        assert(r._2)
      }
    }
  }

  "Selecting 2 columns" should "work fine" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
         Primitives.select(_.pkey, _.long).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 2 columns" should "work fine with Twitter Futures" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).execute() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long).where(_.pkey eqs row.pkey).get()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 3 columns" should "work fine" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 3 columns" should "work fine with Twitter Futures" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).execute() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean).where(_.pkey eqs row.pkey).get()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 4 columns" should "work fine" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 4 columns" should "work fine with Twitter Futures" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).execute() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal).where(_.pkey eqs row.pkey).get()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }


  "Selecting 5 columns" should "work fine" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 5 columns" should "work fine with Twitter Futures" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).execute() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double).where(_.pkey eqs row.pkey).get()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 6 columns" should "work fine" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double, row.float)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double, _.float).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 6 columns" should "work fine with Twitter Futures" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double, row.float)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).execute() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double, _.float).where(_.pkey eqs row.pkey).get()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 7 columns" should "work fine" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double, row.float, row.inet)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double, _.float, _.inet).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 7 columns" should "work fine with Twitter Futures" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double, row.float, row.inet)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).execute() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double, _.float, _.inet).where(_.pkey eqs row.pkey).get()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 8 columns" should "work fine" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double, row.float, row.inet, row.int)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double, _.float, _.inet, _.int).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }

  "Selecting 8 columns" should "work fine with Twitter Futures" in {
    val row = Primitive.sample
    val expected = (row.pkey, row.long, row.boolean, row.bDecimal, row.double, row.float, row.inet, row.int)
    val rcp =  Primitives.insert
      .value(_.pkey, row.pkey)
      .value(_.long, row.long)
      .value(_.boolean, row.boolean)
      .value(_.bDecimal, row.bDecimal)
      .value(_.double, row.double)
      .value(_.float, row.float)
      .value(_.inet, row.inet)
      .value(_.int, row.int)
      .value(_.date, row.date)
      .value(_.uuid, row.uuid)
      .value(_.bi, row.bi).future() flatMap {
      _ => {
        Primitives.select(_.pkey, _.long, _.boolean, _.bDecimal, _.double, _.float, _.inet, _.int).where(_.pkey eqs row.pkey).one()
      }
    }

    rcp successful {
      r => {
        r.isDefined shouldBe true
        r.get shouldBe expected
      }
    }
  }
}
